#include <strings.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include "defines.h"
#include "config.h"
#include "extern.h"

int start_daemon_mode(void) {

	int sock=0;

	init_workerinfo();
	start_logs();
	setup_sighandler();
	sock=setup_port();
	// open_dbs();
	background();
	start_listener(sock);

}

void init_workerinfo(void) {
	
	int i;

	config->thread=(struct child_struct *)calloc(sizeof(struct child_struct), 1);
	config->workers=(struct child_struct **)calloc(sizeof(struct child_struct *), config->children);
	for (i=0; i<config->children; i++) {
		config->workers[i]=(struct child_struct *)calloc(sizeof(struct child_struct), 1);
	}
}

void init_threadinfo(void) {

	config->thread=(struct child_struct *)calloc(sizeof(struct child_struct), 1);



}

int start_logs(void) {

	int out;
	int err;
	char msg[256];
	
	/* as the parent I open the logs */
	
	if (!config->file->logfile || !config->file->errfile) {
		/* This is split over 4 statements just to make the code neater */
		fprintf(stderr, "Logfile(s) not configured.\n");
		fprintf(stderr, "Not using either, will write to stdout & stderr");
		fprintf(stderr, " if you wish to use log files please define names");
		fprintf(stderr, " for _both_ errors and info.\n");
	}
	config->file->logfp=freopen(config->file->logfile, "w+", stdout);
	if (!config->file->logfp) {
		fprintf(stderr, "Error opening logfile: %s ", config->file->logfile);
		perror("");
		fprintf(stderr, "Writing to stdout\n");
	}
	sprintf(msg, "Starting General Logging\n");
	fwrite(msg, sizeof(char), strlen(msg), config->file->logfp);
	config->file->errfp=freopen(config->file->errfile, "w+", stderr);
	if (!config->file->errfp) {
		fprintf(stderr, "Error opening logfile: %s ", config->file->errfile);
		perror("");
		fprintf(stderr, "Writing to stderr\n");
	}
	sprintf(msg, "Starting Error Logging\n");
	fwrite(msg, sizeof(char), strlen(msg), config->file->errfp);
	fflush(stdout);
	fflush(stderr);

}

int setup_sighandler(void) {

	struct sigaction sigact;

	sigact.sa_handler = SIG_IGN;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = 0;
	if (sigaction(SIGTERM, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGTERM failed.\n");
	if (sigaction(SIGHUP, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGHUP failed.\n");
	// if (sigaction(SIGSEGV, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGSEGV failed.\n");

	sigact.sa_handler = do_signal;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags |= SA_RESTART;
	// if (sigaction(SIGSEGV, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGSEGV failed.\n");
	if (sigaction(SIGTERM, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGTERM failed.\n");
	if (sigaction(SIGHUP, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGHUP failed.\n");
	if (sigaction(SIGCHLD, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGHUP failed.\n");

 	// signal(SIGCHLD, catch_child);
	// 	signal(SIGTERM, do_sigterm);
}

void do_signal(int sig) {
	
	switch (sig) {
		case SIGSEGV:
		case SIGTERM:
			if (config->thread->id) printf("Thread[%i]: Caught %s.  Exiting....\n", config->thread->id, sig==SIGSEGV?"SIGSEGV":"SIGTERM");
			kill_workers();
			cleanup();
			exit(0);
			break; // we'll never hit this!
		case SIGCHLD:
			catch_child(sig);
			break;
		default:
			printf("Ignoring Signal [%d]\n", sig);
			break;
	}

}

void kill_workers(void) {

	int i;
	char msg[256];

	if (!allow_kill) {
		if (debug) print_debug(1, "Not sending kill to children\n");
		return;
	}

	for(i=0; i<config->children; i++) {

		if (fcntl(config->workers[i]->parentfd, F_SETFL, 0) 
				|| fcntl(config->workers[i]->childfd, F_SETFL, 0)) {
			fprintf(stderr, "Problem setting pipe for thread %i to blocking\n", i);
			perror("Error");
		}

		memset(msg, '\0', 256);
		sprintf(msg, "K %i", i);
		printf("Sending Kill command to worker[%i]\n", i);
		write(config->workers[i]->childfd, msg, strlen(msg)+1);
		usleep(1);
		memset(msg, '\0', 256);
		sleep_check(config->workers[i]->parentfd, 5);
		read(config->workers[i]->parentfd, msg, 256);

		if (i) printf("Worker[%i]: returned: %s\n", i, msg);

	}
}

void catch_child(int sig) {

	int cstatus;

	wait(&cstatus);
	if (debug) print_debug(3, "Child process exited...\n");
}

void start_threads(void) {

	printf("Threads Configured, but compile time options have not selected threads!\n");

}

void background(void) {

	int i;
	int pfds[2];
	int cfds[2];

	if (!fork() == 0)
		exit(0); /* parent exits - child detaches to the background */
	if (config->threads) start_threads();
	else goto start_workers;

	return;

start_workers:

	for(i=0; i<config->children; i++) {
		config->thread->status=FREE;
		if (pipe(cfds)==-1) {
			perror("background():pipe()");
			cleanup();
			exit(1);
		}
		if (pipe(pfds)==-1) {
			perror("background():pipe()");
			cleanup();
			exit(1);
		}
		if (!fork()) {
			/* Child */
			config->thread->id=i;
			config->thread->parentfd=pfds[1];
			config->thread->childfd=cfds[0];
			close(pfds[0]); /* close the unused end */
			close(cfds[1]); /* close the unused end */

			/* I'm a child - I don't need to know about other children */
			free_workers(config->children);

			do_child();
		} else {
			/* Parent */
			config->workers[i]->id=i;
			config->workers[i]->parentfd=pfds[0];
			config->workers[i]->childfd=cfds[1];
			close(pfds[1]); /* close the unused end */
			close(cfds[0]); /* close the unused end */
		}
	}	
	return;

}

void free_workers(int n) {

	int i;

	for (i=0; i<n; i++) {
		free(config->workers[i]);
	}
	free(config->workers);
	config->workers=NULL;
}

void do_child(void) {

	/* this is the child routing and will accept commands
	 * from the parent, commands will be in the form of 
	 * hosts, ports and types to test.
	 */

	/* set the pipe to non-blocking */
	if (fcntl(config->thread->parentfd, F_SETFL, O_NONBLOCK) 
			|| fcntl(config->thread->childfd, F_SETFL, O_NONBLOCK)) {
		fprintf(stderr, "Thread[%i]: ", config->thread->id);
		perror("Error setting pipes to and from parent to non-blocking");
	}

	config->thread->type=IDLE;
	
	while (1) {
		check_parent(config->thread->parentfd, config->thread->childfd);
		if (config->thread->status == CONNECTING || config->thread->status == FREE) {
			switch (config->thread->type) {
				case SOCKS:
					config->thread->result=test_sock(config->thread->hostname, config->thread->port);
					if (debug) print_debug(2, "Test completed for %s:%d result: %s\n",
							config->thread->hostname, config->thread->port, config->thread->result?
							"NOT OK":"OK");
					config->thread->status=OK;
					break;
				case HTTP:
					config->thread->result=test_proxy(config->thread->hostname, config->thread->port);
					if (debug) print_debug(2, "Test completed for %s:%d result: %s\n",
							config->thread->hostname, config->thread->port, config->thread->result?
							"NOT OK":"OK");
					config->thread->status=OK;
					break;
				case SMTP:
					config->thread->result=test_relay(config->thread->hostname);
					if (debug) print_debug(2, "Test completed for %s:%d result: %s\n",
							config->thread->hostname, config->thread->port, config->thread->result?
							"NOT OK":"OK");
					config->thread->status=OK;
					break;
				case IDLE:
					config->thread->type=IDLE;
					config->thread->status=FREE;
					config->thread->result=OK;
				default:
			}
		sleep_check(config->thread->childfd, SLEEP);
		}
			
	}

}

void sleep_check(int sd, int time) {

	fd_set rfds;
	struct timeval tv;
	int retval;

	/* Watch stdin (fd 0) to see when it has input. */
	FD_ZERO(&rfds);
	FD_SET(sd, &rfds);
	/* Wait up to five seconds. */
	tv.tv_sec = time;
	tv.tv_usec = 0;
	
	retval = select(sd+1, &rfds, NULL, NULL, &tv);
	/* Don't rely on the value of tv now! */
	
	if (debug) {
		if (retval)
			/* FD_ISSET(0, &rfds) will be true. */
			print_debug(3, "Thread[%d]: Parent Pipe has incoming Data\n", config->thread->id);
		else
			print_debug(3, "Thread[%d]: Parent Pipe has No data within %d seconds.\n", config->thread->id, time);
	}
	
}

void check_parent(int psd, int csd) {

	char buf[256];
	int command=0;

	char hostname[256];
	int port=0;
	int type=0;
	int len;

	int r=0;
	int tmp=0;

	char p[1];

	memset(buf, '\0', 256);
	r=read(csd, buf, 256);

	if (r != -1) {
		if (debug) print_debug(4, "Thread[%d]: Received %s from Parent via Pipe (%d bytes)\n", config->thread->id, buf, r);
		switch(*buf) {
			case 'S': /* this is a status request */
				/* Format: 'S' */
				/* Reply: 'S <id> <status>' */
				memset(buf, '\0', 256);
				sprintf(buf, "S %i %i", config->thread->id, config->thread->status);
				write(psd, buf, strlen(buf)+1);
				break;
			case 'T': /* this is a start command */
				/* Formet: 'T <type> <hostname> <port>' */
				/* Reply: 'T <id> <ok>' */
				if (config->thread->status!=FREE) {
					memset(buf, '\0', 256);
					sprintf(buf, "T %i 0", config->thread->id);
					write(psd, buf, strlen(buf)+1);
					if (debug) print_debug(5, "Thread[%d]: Rejecting T Request (NOT FREE)\n", config->thread->id);
					break;
				}
				memset(hostname, '\0', 256);
				sscanf(buf, "%c %i %s %i", p, &type, hostname, &port);

				config->thread->hostname=(char *)calloc(sizeof(char), strlen(hostname)+1);
				strncpy(config->thread->hostname, hostname, strlen(hostname));
				config->thread->port=port;

				memset(buf, '\0', 256);
				sprintf(buf, "T %i 1", config->thread->id, 1);

				switch (type) {
					case HTTP:
						if (debug) print_debug(4, "Thread[%d]: Testing HTTP at: %s:%i\n", config->thread->id, hostname, port);
						config->thread->type=HTTP;
						config->thread->status=CONNECTING;
						break;
					case SMTP:
						if (debug) print_debug(4, "Thread[%d]: Testing SMTP at: %s:%i\n", config->thread->id, hostname, port);
						config->thread->type=SMTP;
						config->thread->status=CONNECTING;
						break;
					case SOCKS:
						if (debug) print_debug(4, "Thread[%d]: Testing SOCKS at: %s:%i\n", config->thread->id, hostname, port);
						config->thread->type=SOCKS;
						config->thread->status=CONNECTING;
						break;
					case IDLE:
						if (debug) print_debug(4, "Thread[%d]: Setting IDLE\n", config->thread->id);
						config->thread->type=IDLE;
						config->thread->status=FREE;
						if (config->thread->hostname) {
							free(config->thread->hostname);
							config->thread->hostname=NULL;
						}
						config->thread->port=0;
						break;
					default:
						fprintf(stderr, "Thread[%i]: Unknown type from parent! [%c]\n", config->thread->id, type);
						sprintf(buf, "T %i 0", config->thread->id);
				}

				write(psd, buf, strlen(buf)+1);
				break;
			case 'R': /* This requests the result */
				/* Format: 'R' */
				/* Reply: R <type> <pass/fail> */
				memset(buf, '\0', 256);
				sprintf(buf, "R %i %i", config->thread->type, config->thread->result);
				write(psd, buf, strlen(buf)+1);
				config->thread->status=FREE; /* reset the thread's status when we give up a result */
				config->thread->type=IDLE;
				break;
			case 'K': /* this is the 'kill' request */
				/* Format: 'K <id>' id must match current ID or the process will ignore the request */
				/* Reply: 'K <id>' */
				sscanf(buf, "%c %i", p, &tmp);
				if (tmp==config->thread->id) {
					/* do die */
					fprintf(stderr, "Thread[%i]: Received KILL request.\n", config->thread->id);
					memset(buf, '\0', 256);
					sprintf(buf, "K %i", config->thread->id);
					write(psd, buf, strlen(buf)+1);
					cleanup();
					exit(0);
				}
				fprintf(stderr, "Thread[%i]: Received KILL request for thread: %i\n", config->thread->id, tmp);
				break;
			default:
				fprintf(stderr, "Thread[%i]: Unknown function [%c] %s\n", config->thread->id, *(buf), buf);
				break;
		}
	}
}
