#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>

#include "extern.h"
#include "dbconfig.h"
#include "defines.h"

int fork_db_thread(void);
void open_dbs(void);
void openlogs(void);

int debug=0;
int verbose=0;

int main(int argc, char **argv) {

	int i;
	int r;
	int f;
	int socketd;

	f=0;

	if (parse_config(argc, argv)) {
		fprintf(stderr, "Unable to parse config file, exiting...\n");
		exit(-1);
	}
	setup_sighandler();
	f=fork();
	if (!f) {
		/* child */
		sleep(1); /* sleep 1 seconds to parent process to clode files and exit */
	} else {
		/* parent (main) thread */
		// close_local_dbs(); /* parent doesn't need to access the db */
		cleanup();
		if (f<0) {
			fprintf(stderr, "Fatal: Unable to fork()\n");
			exit(1);
		}
		fprintf(stdout, "Running Background [%i]\n", f);
		exit(0);
	}
	
	/* we are the child here */

	openlogs();
	open_dbs();
	start_threads();
	socketd=setup_port();
	start_listener(socketd);
	
	exit(10);

}

void openlogs(void) {

	if (!(config->logfile && config->errfile)) {
		fprintf(stderr, "You must define logfiles!\n");
		cleanup();
		exit(1);
	}
	config->log=freopen(config->logfile, "w+", stdout);
	if (!config->log) {
		perror("Error opening logfile");

	}
	config->err=freopen(config->errfile, "w+", stderr);
	if (!config->err) {
		perror("Error opeing error log");
	}

	if (!config->log || !config->err) {
		cleanup();
		exit(1);
	}

}

void open_dbs(void) {

	switch (config->dbtype) {
		case LOCALDB:
			open_local_dbs();
			break;
		case MYSQL:
	//		open_mysql_dbs();
			break;
		case BERKELEY:

			break;
		case ORACLE:

			break;
		default:
			break;

	}

}

void do_dbadd(int id) {

	switch (config->dbtype) {
		case LOCALDB:
			add_local_dbs(id);
			break;
		case MYSQL:
	//		add_mysql_dbs();
			break;
		case BERKELEY:

			break;
		case ORACLE:

			break;
		default:
			break;

	}

}

void do_dbdel(int id) {

	switch (config->dbtype) {
		case LOCALDB:
			open_local_dbs(id);
			break;
		case MYSQL:
	//		del_mysql_dbs();
			break;
		case BERKELEY:

			break;
		case ORACLE:

			break;
		default:
			break;

	}

}

int do_dbcheck(int id) {

	switch (config->dbtype) {
		case LOCALDB:
			return(check_local_dbs(id));
			break;
		case MYSQL:
	//		return(check_mysql_dbs(connection[id]));
			break;
		case BERKELEY:

			break;
		case ORACLE:

			break;
		default:
			return -1;
			break;

	}
}
