#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "extern.h"
#include "defines.h"
#include "dbconfig.h"

/* static locks for locking the db's before writing to them */
static pthread_mutex_t http_lock=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t smtp_lock=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t socks_lock=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t trap_lock=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t white_lock=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t local_lock=PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t oklist_lock=PTHREAD_MUTEX_INITIALIZER;

struct dbreq_struct req[MAX_REQUESTS];

void get_file_lock(int);
void drop_file_lock(int);

void open_local_dbs(void) {

	int r=0;

	dbinfo=(struct db_struct *)calloc(sizeof(struct db_struct), 1);

	if (debug) print_debug(5, "Opening SpamTrap users list file (%s)\n", config->trap);
	if (open_local_db(&dbinfo->trapfd, config->trap)) r=1;
	if (debug) print_debug(5, "Opening bad Reverse DNS list file (%s)\n", config->rdns);
	if (open_local_db(&dbinfo->rdnsfd, config->rdns)) r=1;
	if (debug) print_debug(5, "Opening open-relay list file (%s)\n", config->smtp);
	if (open_local_db(&dbinfo->relayfd, config->smtp)) r=1;
	if (debug) print_debug(5, "Opening http proxy list file (%s)\n", config->http);
	if (open_local_db(&dbinfo->proxyfd, config->http)) r=1;
	if (debug) print_debug(5, "Opening socks list file (%s)\n", config->sock);
	if (open_local_db(&dbinfo->socksfd, config->sock)) r=1;
	if (debug) print_debug(5, "Opening white list file (%s)\n", config->white);
	if (open_local_db(&dbinfo->whitefd, config->white)) r=1;
	if (debug) print_debug(5, "Opening local_db file (%s)\n", config->oklist);
	if (open_local_db(&dbinfo->oklistfd, config->oklist)) r=1;

	if (r) {
		cleanup();
		exit(1);
	}

}

void close_local_dbs(void) {

	close(dbinfo->trapfd);
	close(dbinfo->rdnsfd);
	close(dbinfo->relayfd);
	close(dbinfo->proxyfd);
	close(dbinfo->socksfd);
	close(dbinfo->whitefd);
	close(dbinfo->oklistfd);

	free(dbinfo);
	dbinfo=NULL;
}

int open_local_db(int *fd, char *filename) {


	*fd=open(filename, O_CREAT | O_APPEND | O_RDWR, S_IRUSR | S_IWUSR);
	if (*fd<0) {
		fprintf(stderr, "[%s] ", filename);
		perror("Error opening file");
		return ERROR;
	}
	return 0;

}

int add_local_dbs(int id) {

	int fd;
	int ret;
	
	switch(req[id].client->query) {
		case HTTP:
			if (debug) print_debug(2, "Thread[%d]: adding to the HTTP list\n", id);
			fd=dbinfo->proxyfd;
			break;
		case SMTP:
			if (debug) print_debug(2, "Thread[%d]: adding to the SMTP list\n", id);
			fd=dbinfo->relayfd;
			break;
		case SOCKS:
			if (debug) print_debug(2, "Thread[%d]: adding to the SOCKS list\n", id);
			fd=dbinfo->socksfd;
			break;
		case TRAP:
			if (debug) print_debug(2, "Thread[%d]: adding to the TRAP list\n", id);
			fd=dbinfo->trapfd;
			break;
		case WHITE:
			if (debug) print_debug(2, "Thread[%d]: adding to the WHITE list\n", id);
			fd=dbinfo->whitefd;
			break;
		case OKLIST:
			if (debug) print_debug(2, "Thread[%d]: adding to the OK list\n", id);
			fd=dbinfo->oklistfd;
			break;
		case RDNS:
			if (debug) print_debug(2, "Thread[%d]: adding to the RDNS list\n", id);
			fd=dbinfo->rdnsfd;
			break;
		default:
			if (debug) print_debug(2, "Thread[%d]: Unknown list type %d\n", id);
			return ERROR;
	}
	add_to_file(fd, req[id].query, req[id].client->ipaddr);

}

int del_local_dbs(int id) {

	int fd;
	int ret;
	
	switch(req[id].client->query) {
		case HTTP:
			if (debug) print_debug(2, "Thread[%d]: deleting from the HTTP list\n", id);
			fd=dbinfo->proxyfd;
			break;
		case SMTP:
			if (debug) print_debug(2, "Thread[%d]: deleting from the SMTP list\n", id);
			fd=dbinfo->relayfd;
			break;
		case SOCKS:
			if (debug) print_debug(2, "Thread[%d]: deleting from the SOCKS list\n", id);
			fd=dbinfo->socksfd;
			break;
		case TRAP:
			if (debug) print_debug(2, "Thread[%d]: deleting from the TRAP list\n", id);
			fd=dbinfo->trapfd;
			break;
		case WHITE:
			if (debug) print_debug(2, "Thread[%d]: deleting from the WHITE list\n", id);
			fd=dbinfo->whitefd;
			break;
		case OKLIST:
			if (debug) print_debug(2, "Thread[%d]: deleting from the OK list\n", id);
			fd=dbinfo->oklistfd;
			break;
		case RDNS:
			if (debug) print_debug(2, "Thread[%d]: deleting from the RDNS list\n", id);
			fd=dbinfo->rdnsfd;
			break;
		default:
			if (debug) print_debug(2, "Thread[%d]: Unknown list type %d\n", id);
			return ERROR;
	}
	delete_from_file(fd, req[id].query, req[id].client->ipaddr);

}

int check_local_dbs(int id) {

	int fd;
	int ret;
	
	switch(req[id].query) {
		case HTTP:
			if (debug) print_debug(2, "Thread[%d]: Checking HTTP list\n", id);
			fd=dbinfo->proxyfd;
			break;
		case SMTP:
			if (debug) print_debug(2, "Thread[%d]: Checking SMTP list\n", id);
			fd=dbinfo->relayfd;
			break;
		case SOCKS:
			if (debug) print_debug(2, "Thread[%d]: Checking SOCKS list\n", id);
			fd=dbinfo->socksfd;
			break;
		case TRAP:
			if (debug) print_debug(2, "Thread[%d]: Checking TRAP list\n", id);
			fd=dbinfo->trapfd;
			break;
		case WHITE:
			if (debug) print_debug(2, "Thread[%d]: Checking WHITE list\n", id);
			fd=dbinfo->whitefd;
			break;
		case OKLIST:
			if (debug) print_debug(2, "Thread[%d]: Checking OK list\n", id);
			fd=dbinfo->oklistfd;
			break;
		case RDNS:
			if (debug) print_debug(2, "Thread[%d]: Checking RDNS list\n", id);
			fd=dbinfo->rdnsfd;
			break;
		default:
			if (debug) print_debug(2, "Thread[%d]: Unknown list type %d\n", id);
			return ERROR;
	}
	
	ret=scan_file(fd, req[id].query, req[id].client->ipaddr);
	return ret;
}

int add_to_file(int fd, int type, char *ipaddr) {

	char buf[256];

	memset(buf, '\0', 256);
	sprintf(buf, "%s\n", ipaddr);

	get_file_lock(type);
	lseek(fd, 0, SEEK_END);
	write(fd, buf, strlen(buf));
	fsync(fd); /* this is used instead of fdatasync() deliberately */
	drop_file_lock(type);

}

int delete_from_file(int fd, int type, char *ipaddr) {

	/* deliberately left - this is for the future */

}

int scan_file(int fd, int type, char *ipaddr) {

	
	char *buf, *file;
	char *line, *p;
	int len;

	len=strlen(ipaddr);

	lseek(fd, 0, SEEK_SET);
	buf=(char *)calloc(sizeof(char), FILE_BUF);
	
	get_file_lock(type);
	file=line=read_file(fd);
	drop_file_lock(type);

	if (debug) print_debug(5, "Searching for: %s\n", ipaddr);

	while (line && *line && *(line+1) && *(line+2)) {
		p=line;
		while (*p && *p != '\n' && *p != '\r' && *p != '\t' && *p != ' ') p++;
		if (*p) {
			*p++='\0';
			while (*p && *p == '\n' && *p == '\r' && *p == '\t' && *p == ' ') p++;
		} else {
			p=NULL;
		}

		if (p && *line==*ipaddr && *(line+1) == *(ipaddr+1) && *(line+2) == *(ipaddr+2)) {
			memset(buf, '\0', FILE_BUF);
			strncpy(buf, line, (p-line)<FILE_BUF?(size_t)(p-line):(size_t)FILE_BUF-1);
			if (debug) print_debug(5, "Attempting match on: %s ", buf);
			if (*buf != '#')  {
				if (!strncmp(line, ipaddr, len)) {
					if (debug) printf("matched!\n");
					/* we have a match! */
					free(buf);
					free(file);
					return MATCHED;
				}
			if (debug) printf("not matched.\n");
			}
			p++;
		}
		line=p;
	}
	free(buf);
	free(file);
	return NOT_MATCHED;
}

void get_file_lock(int type) {

	switch (type) {
		case HTTP:
			pthread_mutex_lock(&http_lock);
			break;
		case SMTP:
			pthread_mutex_lock(&smtp_lock);
			break;
		case SOCKS:
			pthread_mutex_lock(&socks_lock);
			break;
		case TRAP:
			pthread_mutex_lock(&trap_lock);
			break;
		case WHITE:
			pthread_mutex_lock(&white_lock);
			break;
		case LOCALLIST:
			pthread_mutex_lock(&local_lock);
			break;
		case OKLIST:
			pthread_mutex_lock(&oklist_lock);
			break;
	}

}

void drop_file_lock(int type) {

	switch (type) {
		case HTTP:
			pthread_mutex_unlock(&http_lock);
			break;
		case SMTP:
			pthread_mutex_unlock(&smtp_lock);
			break;
		case SOCKS:
			pthread_mutex_unlock(&socks_lock);
			break;
		case TRAP:
			pthread_mutex_unlock(&trap_lock);
			break;
		case WHITE:
			pthread_mutex_unlock(&white_lock);
			break;
		case LOCALLIST:
			pthread_mutex_unlock(&local_lock);
			break;
		case OKLIST:
			pthread_mutex_unlock(&oklist_lock);
			break;
	}
}
