#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>

#include "extern.h"
#include "config.h"
#include "defines.h"

int test_relay (char *hostname) {

	char *buf, *mp;
	char msg[256];
	char tmsg[256];
	char mydomain[256];
	char myhost[256];
	int sockfd;
	int res, *resp;
	int ret;

	if (gethostname(myhost,256)) sprintf(myhost, "nowhere");
	if (getdomainname(mydomain,256)) sprintf(mydomain, "example.com");


	res=0;
	resp=&res;

	sockfd=net_connect(hostname, 25);

	if (!sockfd && !config->level) { /* Paranoid mode - reject if there is no remote smtp server */
		close(sockfd);
		if (config->level) reject(1);
		else config->thread->result=1;
		return 1;
	}

	if (!sockfd) return 0;

	mp=buf=net_read(sockfd);

	if (debug) print_debug(4, "Server returned: %s\n", buf);
	
	res=(*buf=='2');
	free(mp);
	if (!res) {
		printf("Got bad result from SMTP server: %i\n", res);
		sprintf(msg, "QUIT\r\n");
		net_send(sockfd, msg);
		if (config->strictsmtp) {
			return 1;
		} else {
			return 0;
		}
	}
			
	sprintf(msg, "HELO %s\r\n", myhost);
	res=do_smtp_proto(sockfd, msg, '2');
	if (!res && config->strictsmtp) {
		ret=1;
		goto quit;
	}

	sprintf(msg, "MAIL FROM: sorbs-test@%s\r\n", mydomain);
	res=do_smtp_proto(sockfd, msg, '2');
	if (!res) {
		ret=1; /* if our MAIL FROM is rejected this server is probably expecting us */
		goto quit;
	}

	sprintf(msg, "RCPT TO: sorbs-test@%s\r\n", mydomain);
	res=do_smtp_proto(sockfd, msg, '5');
	if (!res) goto relaycheck;

	sprintf(msg, "RCPT TO: sorbs-test%%%s@%s\r\n", mydomain, hostname);
	res=do_smtp_proto(sockfd, msg, '5');
	if (!res) goto relaycheck;

	sprintf(msg, "RCPT TO: <@%s:sorbs-test@%s>\r\n", hostname, mydomain);
	res=do_smtp_proto(sockfd, msg, '5');
	if (!res) goto relaycheck;

	sprintf(msg, "RCPT TO: <@%s:sorbs-test%%%s@%s>\r\n", hostname, mydomain, hostname);
	res=do_smtp_proto(sockfd, msg, '5');
	if (!res) goto relaycheck;

	sprintf(msg, "RSET\r\n");
	res=do_smtp_proto(sockfd, msg, '2');
	if (!res && config->strictsmtp) {
		ret=1;
		goto quit;
	}

	sprintf(msg, "MAIL FROM: sorbs-test@%s\r\n", hostname);
	res=do_smtp_proto(sockfd, msg, '2');
	if (!res) return 1; /* if our MAIL FROM is rejected this server is probably expecting us */

	sprintf(msg, "RCPT TO: sorbs-test@%s\r\n", mydomain);
	res=do_smtp_proto(sockfd, msg, '5');
	if (!res) goto relaycheck;

	sprintf(msg, "RCPT TO: sorbs-test%%%s@%s\r\n", mydomain, hostname);
	res=do_smtp_proto(sockfd, msg, '5');
	if (!res) goto relaycheck;

	sprintf(msg, "RCPT TO: <@%s:sorbs-test@%s>\r\n", hostname, mydomain);
	res=do_smtp_proto(sockfd, msg, '5');
	if (!res) goto relaycheck;

	sprintf(msg, "RCPT TO: <@%s:sorbs-test%%%s@%s>\r\n", hostname, mydomain, hostname);
	res=do_smtp_proto(sockfd, msg, '5');
	if (!res) goto relaycheck;

	sprintf(tmsg, "RSET\r\n");
	do_smtp_proto(sockfd, tmsg, '2');

	goto quit;

relaycheck:

	sprintf(tmsg, "RSET\r\n");
	do_smtp_proto(sockfd, tmsg, '2');
	sprintf(tmsg, "MAIL FROM: sorbs-test@%s\r\n", mydomain);
	do_smtp_proto(sockfd, tmsg, '2');

	res=do_smtp_proto(sockfd, msg, '2');
	if (res || (!res && config->strictsmtp)) ret=1;

quit:

	sprintf(msg, "QUIT\r\n");
	res=do_smtp_proto(sockfd, msg, '2');
	close(sockfd);

	return ret;
}

int do_smtp_proto(int sockfd, char *msg, int ok) {

	char *mp, *buf;
	char c;
	int res;

	res=0;

	net_send(sockfd, msg);
	mp=buf=net_read(sockfd);
	if (debug) print_debug(4, "Server returned: %s\n", buf);
	res=(*buf==ok);
	free(mp);
	if (!res) {
		if (debug) print_debug(3, "Got bad result from SMTP server: %c%c%c\n", *buf, *(buf+1), *(buf+2));
	}
	return res;
}

